#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>
#include <regex.h>

#include "config.h"
#include "imapfilter.h"
#include "connect.h"
#include "log.h"
#include "imap.h"


int sock;
extern account_data account;
extern int options;


/*
 * Connects to mailserver.
 */
int establish_connection()
{
    struct sockaddr_in name;
    struct hostent *hostinfo;

    memset((char *) &name, 0, sizeof(struct sockaddr_in));

    sock = socket(PF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	error("imapfilter: create socket; %s\n", strerror(errno));
	return FAILURE;
    }

    if (!(hostinfo = gethostbyname(account.server))) {
	error("imapfilter: get network host entry; %s\n", strerror(errno));
	close_connection();
	return FAILURE;
    }

    name.sin_family = AF_INET;
    name.sin_port = htons(account.port);
    name.sin_addr = *(struct in_addr *) hostinfo->h_addr;

    if (connect(sock, (struct sockaddr *) &name, sizeof(struct sockaddr))) {
	error("imapfilter: initiating connection; %s\n", strerror(errno));
	close_connection();
	return FAILURE;
    }

    verbose("Connected to %s.\n", hostinfo->h_name, account.port);

    return SUCCESS;
}


/*
 * Disconnects from mailserver.
 */
int close_connection(void)
{
    if (close(sock)) {
	error("imapfilter: closing socket; %s\n", strerror(errno));
	return FAILURE;
    } else {
	return SUCCESS;
    }
}


/*
 * Reads (if any) data the server sent.
 */
int get_response(char *buf)
{
    int flags;
    fd_set rfds;
    struct timeval tv;

    memset(buf, 0, RESPONSE_BUFFER_MAX);

    tv.tv_sec = 10;
    tv.tv_usec = 0;

    flags = fcntl(sock, F_GETFL, 0);
    fcntl(sock, F_SETFL, flags | O_NONBLOCK);

    FD_ZERO(&rfds);
    FD_SET(sock, &rfds);

    if ((select(sock + 1, &rfds, NULL, NULL, &tv) > 0)
	&& FD_ISSET(sock, &rfds)) {
	if (read(sock, buf, RESPONSE_BUFFER_MAX - 1) != -1) {
#ifdef DEBUG
	    printf("\n%s\n", buf);
#endif
	    if (options & OPTION_DETAILS_VERBOSE)
		analyze_response(buf);

	    fcntl(sock, F_SETFL, flags);
	    return SUCCESS;
	}
    }

    fcntl(sock, F_SETFL, flags);

    return FAILURE;
}


/*
 * Clears stream from any data server sent without processing them.
 */
int clear_stream(void)
{
    int r;
    unsigned int saveopt;
    char buf[RESPONSE_BUFFER_MAX];

    saveopt = options;		/* Needed so response won't be analyzed. */
    options &= OPTION_DETAILS_CLEAR;

    r = get_response(buf);

    options = saveopt;

    return r;
}


/*
 * Processes the data that server sent due to IMAP SEARCH client request.
 */
int search_response(char *results)
{
    int s, len;
    char buf[RESPONSE_BUFFER_MAX];
    regex_t compexp;
    regmatch_t match[2];
    const char *reg = "\\* SEARCH([[:digit:] ]*)";

    len = strlen(results);

    get_response(buf);

    regcomp(&compexp, reg, REG_EXTENDED);

    if (!regexec(&compexp, buf, 2, match, 0)) {
	s = min(match[1].rm_eo - match[1].rm_so, SEARCH_RESULTS_MAX - 1);
	strncpy(results + len, buf + match[1].rm_so, s);
	results[s + len] = 0;
	return SUCCESS;
    } else
	return FAILURE;
}


/*
 * Processes the data that server sent due to IMAP FETCH client request.
 */
void fetch_response(char *results)
{
    int i, s, len = 0;
    char buf[RESPONSE_BUFFER_MAX];
    regex_t compexp[3];
    regmatch_t match[1];
    const char *reg[3] = {
	"Date: [[:print:]]*",
	"From: [[:print:]]*",
	"Subject: [[:print:]]*"
    };

    get_response(buf);

    for (i = 0; i < 3; i++) {
	regcomp(&compexp[i], reg[i], REG_EXTENDED);

	if (!regexec(&compexp[i], buf, 1, match, 0)) {
	    s = min(match[0].rm_eo - match[0].rm_so,
		    HEADERS_MAX - len - 1);
	    strncpy(results + len, buf + match[0].rm_so, s);
	    results[s + len] = '\n';
	    results[s + len + 1] = 0;
	    len = strlen(results);
	}
    }
}


/*
 * Checks if response of server to client's request was succesfully
 * delivered or there was some kind of error.
 */
void analyze_response(char *buf)
{
    int s;
    regex_t compexp;
    regmatch_t match[2];
    const char *reg = "[[:xdigit:]]{4,4} ([[:print:]]*)";
    char response[RESPONSE_BUFFER_MAX];

    regcomp(&compexp, reg, REG_EXTENDED);

    if (!regexec(&compexp, buf, 2, match, 0)) {
	s = min(match[1].rm_eo - match[1].rm_so, RESPONSE_BUFFER_MAX - 1);
	strncpy(response, buf + match[1].rm_so, s);
	response[s] = 0;
	printf("Server response: %s\n", response);
    }
}
