#include <stdio.h>
#include <errno.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include "config.h"
#include "imapfilter.h"
#include "file.h"
#include "log.h"


account_data account;
filter_entry **dfilters;	/* Filters of DENY type. */
filter_entry **afilters;	/* Filters of ALLOW type. */
int dfcnt = 0;
int afcnt = 0;			/* DENY and ALLOW type filters found
				   in the configuration file. */
unsigned int dlimit = 0;
unsigned int alimit = 0;	/* DENY and ALLOW message size limits. */


/*
 * Finds path to configuration file, opens it and calls parse_config().
 */
int read_config(char *cfg)
{
    int r;
    FILE *fp;
    char *home = NULL;

    if (!cfg) {
	cfg = (char *) malloc(PATH_MAX * sizeof(char));

	home = getenv("HOME");

	snprintf(cfg, PATH_MAX, "%s/%s", home, ".imapfilterrc");
    }
#ifdef DEBUG
    printf("debug: configuration file: %s\n", cfg);
#endif

    fp = fopen(cfg, "r");

    if (!fp) {
	fprintf(stderr, "imapfilter: Could not open config file %s; %s\n",
		cfg, strerror(errno));
	return FAILURE;
    }

    r = parse_config(fp);

    fclose(fp);

    return r;
}

/*
 * Parses configuration file.
 */
int parse_config(FILE * fp)
{
    int i;
    unsigned int row = 0;
    char line[LINE_MAX];
    regex_t comreg[5];
    regmatch_t match[4];
    const char *reg[5] = {
	"^[[:blank:]]*(SERVER|PORT|USERNAME|PASSWORD|LOGFILE)[[:blank:]]*=[[:blank:]]*([[:graph:]]*)[[:blank:]]*",
	"^[[:blank:]]*(DENY|ALLOW)[[:blank:]]*=[[:blank:]]*(FROM|CC|BCC|SUBJECT|TO):? ([[:print:]]*)",
	"^[[:blank:]]*(DENY|ALLOW)[[:blank:]]*=[[:blank:]]*([[:graph:]]+):? ([[:print:]]*)",
	"^[[:blank:]]*(DENY_LIMIT|ALLOW_LIMIT)[[:blank:]]*=[[:blank:]]*([[:digit:]]*)[[:blank:]]*",
	"^[[:blank:]]*#{0,1}.*"
    };

    for (i = 0; i < 5; i++)
	regcomp(&comreg[i], reg[i], REG_EXTENDED);

    while (fgets(line, LINE_MAX - 1, fp)) {
	row++;
	if (!regexec(&comreg[0], line, 3, match, 0)) {
	    set_account(line, match);
	    continue;
	} else if (!regexec(&comreg[1], line, 4, match, 0)) {
	    standard_filter(line, match);
	    continue;
	} else if (!regexec(&comreg[2], line, 4, match, 0)) {
	    custom_filter(line, match);
	    continue;
	} else if (!regexec(&comreg[3], line, 3, match, 0)) {
	    set_limits(line, match);
	    continue;
	} else if (!regexec(&comreg[4], line, 1, match, 0))
	    continue;
	else {
	    fprintf(stderr,
		    "imapfilter: parse error in config file at row %d\n",
		    row);
	    return FAILURE;
	}
    }

    /* Fail if no filters were defined. */
    if (!dfcnt && !afcnt) {
	fprintf(stderr, "imapfilter: no filters defined in config file\n");
	return FAILURE;
    }
#ifdef DEBUG
    printf("debug: account setting SERVER: '%s'\n"
	   "debug: account setting PORT: %d\n"
	   "debug: account setting USERNAME: '%s'\n"
	   "debug: account setting PASSWORD: '%s'\n"
	   "debug: account setting LOGFILE: '%s'\n"
	   "debug: setting DENY_LIMIT: %d\n"
	   "debug: setting ALLOW_LIMIT: %d\n",
	   account.server, account.port, account.username,
	   account.password, account.logfile, dlimit, alimit);
#endif
    return SUCCESS;
}


/*
 * NULL terminate account settings.
 */
void prepare_account(void)
{
    account.server[0] = 0;
    account.port = 143;
    account.username[0] = 0;
    account.password[0] = 0;
    account.logfile[0] = 0;
}


/*
 * An account setting was found from parse_config(), changes the
 * apropriate variable.
 */
void set_account(char *line, regmatch_t * match)
{
    int s;
    char port[6];

    if (!strncmp(line + match[1].rm_so, "SERVER", 6)) {
	s = min((match[2].rm_eo - match[2].rm_so), SERVER_MAX - 1);
	strncpy(account.server, line + match[2].rm_so, s);
	account.server[s] = 0;
    } else if (!strncmp(line + match[1].rm_so, "PORT", 4)) {
	s = min((match[2].rm_eo - match[2].rm_so), sizeof(port));
	strncpy(port, line + match[2].rm_so, s);
	port[s] = 0;
	account.port = strtoul(port, NULL, 0);
    } else if (!strncmp(line + match[1].rm_so, "USERNAME", 8)) {
	s = min((match[2].rm_eo - match[2].rm_so), USERNAME_MAX - 1);
	strncpy(account.username, line + match[2].rm_so, s);
	account.username[s] = 0;
    } else if (!strncmp(line + match[1].rm_so, "PASSWORD", 8)) {
	s = min((match[2].rm_eo - match[2].rm_so), PASSWORD_MAX - 1);
	strncpy(account.password, line + match[2].rm_so, s);
	account.password[s] = 0;
    } else if (!strncmp(line + match[1].rm_so, "LOGFILE", 7)) {
	s = min((match[2].rm_eo - match[2].rm_so), PATH_MAX - 1);
	strncpy(account.logfile, line + match[2].rm_so, s);
	account.logfile[s] = 0;
    }
}


/*
 * A filter entry was found from parse_config() processes it and saves it.
 */
void set_filters(char *line, regmatch_t * match, int expl)
{
    int s;
    int *fcnt;
    filter_entry ***filters;

    if (!(strncmp(line + match[1].rm_so, "DENY", 4))) {
#ifdef DEBUG
	printf("debug: filter entry: DENY");
#endif
	fcnt = &dfcnt;
	filters = &dfilters;
    } else {
#ifdef DEBUG
	printf("debug: filter entry: ALLOW");
#endif
	fcnt = &afcnt;
	filters = &afilters;
    }

    /* Every FILTER_ENTRIES_MAX, space for more filters is realloc() -ed. */
    if (!*fcnt)
	*filters =
	    (filter_entry **) malloc(FILTER_ENTRIES_MAX *
				     sizeof(filter_entry *));
    else if (!(*fcnt % FILTER_ENTRIES_MAX))
	*filters = (filter_entry **) realloc(*filters,
					     (((*fcnt /
						FILTER_ENTRIES_MAX) +
					       1) * FILTER_ENTRIES_MAX) *
					     sizeof(filter_entry *));

    (*filters)[*fcnt] = (filter_entry *) malloc(sizeof(filter_entry));

    (*filters)[*fcnt]->custom = expl;

    s = min((match[2].rm_eo - match[2].rm_so), FIELD_NAME_MAX - 1);
    strncpy((*filters)[*fcnt]->name, line + match[2].rm_so, s);
    (*filters)[*fcnt]->name[s] = 0;

    s = min((match[3].rm_eo - match[3].rm_so), FIELD_BODY_MAX - 1);
    strncpy((*filters)[*fcnt]->body, line + match[3].rm_so, s);
    (*filters)[*fcnt]->body[s] = 0;

#ifdef DEBUG
    printf(" '%s': '%s'\n", (*filters)[*fcnt]->name,
	   (*filters)[*fcnt]->body);
#endif

    (*fcnt)++;
}


/*
 * Sets the DENY and ALLOW limits from the values found in config file.
 */
void set_limits(char *line, regmatch_t * match)
{
    int s;
    char limit[10];

    s = min((match[2].rm_eo - match[2].rm_so), sizeof(limit));
    strncpy(limit, line + match[2].rm_so, s);
    limit[s] = 0;

    if (!strncmp(line + match[1].rm_so, "DENY_LIMIT", 10)) {
	dlimit = strtoul(limit, NULL, 0);
    } else {
	alimit = strtoul(limit, NULL, 0);
    }
}
