#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "config.h"
#include "imapfilter.h"
#include "imap.h"
#include "connect.h"
#include "log.h"


static int tag = 0xA000;	/* Every IMAP command is prefixed with a
				   unique [:alnum:] string. */
extern int sock;
extern int dfcnt, afcnt;
extern filter_entry **dfilters, **afilters;
extern account_data account;
extern int options;
extern unsigned int dlimit, alimit;


/*
 * Sends to server data; a command.
 */
int send_command(char *command)
{

#ifdef DEBUG
    printf("debug: sending command: %s", command);
#endif

    if (write(sock, command, strlen(command)) == -1) {
	error("imapfilter: error while sending command; %s",
	      strerror(errno));
	return FAILURE;
    }

    return SUCCESS;
}


/*
 * Calls send_command() and get_response().
 */
int send_command_get_response(char *command)
{
    char buf[RESPONSE_BUFFER_MAX];

    if (!send_command(command)) {
	get_response(buf);
	return SUCCESS;
    } else
	return FAILURE;
}


#ifdef DEBUG
/*
 * IMAP NOOP: does nothing always succeeds.
 */
int imap_noop(void)
{
    char command[SMALL_COMMAND_MAX];

    verbose("Client request:  NOOP\n");

    snprintf(command, SMALL_COMMAND_MAX, "%X NOOP\r\n", tag++);

    return send_command_get_response(command);

}
#endif


/*
 * IMAP LOGOUT: informs server that client is done.
 */
int imap_logout(void)
{
    char command[SMALL_COMMAND_MAX];

    verbose("Client request:  LOGOUT\n");

    snprintf(command, SMALL_COMMAND_MAX, "%X LOGOUT\r\n", tag++);

    return send_command_get_response(command);
}


/*
 * IMAP LOGIN: identifies client to server.
 */
int imap_login(void)
{
    char command[MEDIUM_COMMAND_MAX];

    verbose("Client request:  LOGIN\n");

    snprintf(command, MEDIUM_COMMAND_MAX, "%X LOGIN \"%s\" \"%s\"\r\n",
	     tag++, account.username, account.password);

    return send_command_get_response(command);
}


/*
 * IMAP SELECT: selects a mailbox in which messages can be accessed.
 */
int imap_select(void)
{
    char command[SMALL_COMMAND_MAX];

    verbose("Client request:  SELECT\n");

    snprintf(command, SMALL_COMMAND_MAX, "%X SELECT INBOX\r\n", tag++);

    return send_command_get_response(command);
}

/*
 * IMAP SEARCH: searches the mailbox for messages that match certain criteria.
 */
int imap_search(char *results)
{
    char command[BIG_COMMAND_MAX];

    verbose("Client request:  SEARCH\n");

    generate_search_filters(command);
    if (send_command(command))
	return FAILURE;
    search_response(results);

    verbose("Client request:  SEARCH\n");

    if (alimit && afcnt) {
	generate_search_limits(command);
	if (send_command(command))
	    return FAILURE;
	search_response(results);
    }

    return SUCCESS;
}


/*
 * IMAP FETCH: retrieves data associated with a message.
 */
int imap_fetch(unsigned int m, char *results)
{
    char command[MEDIUM_COMMAND_MAX];

    verbose("Client request:  FETCH\n");

    snprintf(command, MEDIUM_COMMAND_MAX,
	     "%X FETCH %d BODY[HEADER.FIELDS (\"DATE\" \"FROM\" \"SUBJECT\")]\r\n",
	     tag++, m);

    if (!send_command(command)) {
	fetch_response(results);
	return SUCCESS;
    } else
	return FAILURE;
}


/*
 * IMAP STORE: alters data associated with a message.
 */
int imap_store(unsigned int m)
{
    char command[MEDIUM_COMMAND_MAX];

    verbose("Client request:  STORE\n");

    snprintf(command, MEDIUM_COMMAND_MAX,
	     "%X STORE %d +FLAGS \\Deleted\r\n", tag++, m);

    return send_command_get_response(command);
}


/*
 * IMAP EXPUNGE: premanently removes any messages with the \Deleted flag set.
 */
int imap_expunge(void)
{
    char command[SMALL_COMMAND_MAX];

    verbose("Client request:  EXPUNGE\n");

    snprintf(command, SMALL_COMMAND_MAX, "%X EXPUNGE\r\n", tag++);

    return send_command_get_response(command);
}


/*
 * Prepare according to the filters defined by the user the IMAP SEARCH
 * command that will be sent.  This command will search for DENY, ALLOW,
 * and DENY_LIMIT filters.
 */
void generate_search_filters(char *command)
{
    int len;

    snprintf(command, BIG_COMMAND_MAX, "%X SEARCH", tag++);


    deny_filters(command);

    if (dlimit) {
	len = strlen(command);
	snprintf(command + len, BIG_COMMAND_MAX - len, " LARGER %d",
		 dlimit);
    } else
	if (!dfcnt) {		/* If no DENY filters were defined, then
				   deny all except the ALLOW filters. */
	    len = strlen(command);
	    snprintf(command + len, BIG_COMMAND_MAX - len, " ALL");
	}

    allow_filters(command, "NOT");

    len = strlen(command);
    snprintf(command + len, BIG_COMMAND_MAX - len, "\r\n");
}


/*
 * Prepare according to the filters defined by the user the IMAP SEARCH
 * command that will be sent.  This command will search for the ALLOW_LIMIT
 * filter.
 */
void generate_search_limits(char *command)
{
    int len;

    snprintf(command, BIG_COMMAND_MAX, "%X SEARCH", tag++);

    allow_filters(command, "OR");

    len = strlen(command);
    snprintf(command + len, BIG_COMMAND_MAX - len, " LARGER %d", alimit);

    len = strlen(command);
    snprintf(command + len, BIG_COMMAND_MAX - len, "\r\n");

}


/*
 * Adds to the IMAP SEARCH command all the DENY type filters.
 */
void deny_filters(char *command)
{
    int f = 0;
    int len;

    if (dfcnt > 0) {
	while (f < (dfcnt - 1)) {
	    len = strlen(command);
	    snprintf(command + len, BIG_COMMAND_MAX - len,
		     " OR %s%s \"%s\"",
		     custom_header(dfilters[f]->custom),
		     dfilters[f]->name, dfilters[f]->body);
	    f++;
	}

	len = strlen(command);
	snprintf(command + len, BIG_COMMAND_MAX - len, " %s%s%s \"%s\"",
		 (dlimit ? "OR " : ""),
		 custom_header(dfilters[f]->custom),
		 dfilters[f]->name, dfilters[f]->body);
    }
}


/*
 * Adds to IMAP SEARCH command all the ALLOW type filters.
 */
void allow_filters(char *command, char *key)
{
    int f = 0;
    int len;

    if (afcnt > 0) {
	while (f < (afcnt - 1)) {
	    len = strlen(command);
	    snprintf(command + len, BIG_COMMAND_MAX - len,
		     " %s %s%s \"%s\"", key,
		     custom_header(afilters[f]->custom),
		     afilters[f]->name, afilters[f]->body);
	    f++;
	}

	len = strlen(command);
	snprintf(command + len, BIG_COMMAND_MAX - len, " %s%s%s \"%s\"",
		 (!strncmp(key, "NOT", 3) ? "NOT " : ""),
		 custom_header(afilters[f]->custom),
		 afilters[f]->name, afilters[f]->body);
    }
}


/*
 * Convert the string with the UID's of the messages to be deleted to
 * integers, send the appropriate command and maybe print some headers
 * of the "to be deleted" message.
 */
void delete_messages(char *unwanted)
{
    unsigned long int messg;
    char headers[HEADERS_MAX];

    do {
	messg = strtoul(unwanted, &unwanted, 0);

	if (options & OPTION_SHOW_HEADERS) {
	    imap_fetch(messg, headers);

	    if (!(options & OPTION_DETAILS_QUITE))
		printf("Deleting message:\n%s", headers);

	    if (!(options & OPTION_TEST_MODE))
		log_info("%s", headers);
	}

	if (!(options & OPTION_TEST_MODE))
	    imap_store(messg);

    } while (*unwanted);
}
